import re
from .logsetup import UserLogger
from .config import dummy_context_mgr
from infinisdk.core.exceptions import APICommandFailed
from distutils.version import LooseVersion
from .infinibox import IMMUTABLE_SUPPORTED_VER
import sys

log = UserLogger(__name__)


def filter_irrelevant_snapshots(snapshots, volume, basename):
    def pred(snapshot):
        name = snapshot.get_name()
        prefix = r'^.*\.{}\.'.format(re.escape(basename))
        date = r'\d{4}\.\d{2}\.\d{2}_\d{6}$'  # yyyy.mm.dd_hhmmss
        return bool(re.match(prefix + date, name))
    return [s for s in snapshots if pred(s)]


def make_snapshot_name(volume, basename, time, uuid):
    from datetime import datetime
    dt = datetime.fromtimestamp(time)
    if uuid == True:
        name = volume.get_id()
    else:
        name = volume.get_name()
    return "{}.{}.{}".format(name, basename, dt.strftime("%Y.%m.%d_%H%M%S"))


def sort_and_filter_snapshots(snapshots, volume, basename):
    """:return: sorted list of snapshots where snapshots
    not conforming to the name template are removed"""
    return sorted(filter_irrelevant_snapshots(snapshots, volume, basename),
                  key=lambda o: o.get_name())


def trim_number_of_snapshots(infinibox, snapshots, volume, basename, retain, dryrun, suppress=False):
    """We try to delete as many snapshots as needed to keep the number of snapshots less or equal to self.retain."""
    snapshots = sort_and_filter_snapshots(snapshots, volume, basename)
    num_to_delete = max(0, len(snapshots) - retain)
    snapshots_to_delete = snapshots[:num_to_delete]
    metadata = list()
    for snap in snapshots_to_delete:
        metadata.append(snap.get_all_metadata())
        name = None
        try:
            name = snap.get_name()
            # only check expiry if the array supports it
            if LooseVersion(infinibox.get_version()) >= IMMUTABLE_SUPPORTED_VER:
                lock_expiry = snap.get_lock_expires_at()
            else:
                lock_expiry = False
            if not dryrun:
                if snap.get_type_name() == 'cons_group':
                    with infinibox.api.added_headers_context({"X-INFINIDAT-SUPPRESS-EVENTS": "true"}) if suppress else dummy_context_mgr():
                        snap.delete(delete_members=True)
                    log.uinfo("Deleted snapshot {}".format(name))
                elif snap.has_children() == True:
                    log.uwarn(
                        "Clone exists for snapshot {}. Not deleting.".format(name))
                elif snap.is_mapped():
                    log.uwarn(
                        "Snapshot {} is mapped. Not deleting.".format(name))
                    # PSDEV: add immutability support
                elif lock_expiry and lock_expiry > lock_expiry.now():
                    log.uwarn("Snap lock for %s will expire %s. Not deleting." %
                              (name, lock_expiry.humanize()))
                else:
                    with infinibox.api.added_headers_context({"X-INFINIDAT-SUPPRESS-EVENTS": "true"}) if suppress else dummy_context_mgr():
                        snap.delete()
                        log.uinfo("Deleted snapshot {}".format(name))
        except APICommandFailed as e:
            if e.error_code in ('NOT_SUPPORTED_IN_SYSTEM_STATE'):
                log.uabort('Operation not supported in STANDBY mode. Exiting.')
            else:
                log.uwarn(
                    "Failed to delete snapshot {}, moving on {}".format(name, repr(e)), exc_info=True)
        except Exception as e:
            log.uwarn(
                "Failed to delete snapshot {}, moving on {}".format(name, repr(e)), exc_info=True)
    return metadata


def create_latest_snapshot(infinibox, volume, basename, dryrun, uuid,
                           suppress=False, lock_days=0):
    """ Create latest snapshot <volname>.<basename>.YYYY.MM.DD_HHMMSS"""
    from time import time
    snapshot = None
    try:
        name = make_snapshot_name(volume, basename, time(), uuid)
        if len(name) > 64:
            log.uwarn('Snapshot name too long: {}'.format(name))
            return None
        voltype = volume.get_type_name()
        log.uinfo(
            "Creating snapshot {} for {} '{}'".format(name, voltype, volume.get_name()))
        if dryrun:
            return None
        with infinibox.api.added_headers_context({"X-INFINIDAT-SUPPRESS-EVENTS": "true"}) if suppress else dummy_context_mgr():
            # PSDEV: add immutability support
            snapshot = volume.create_snapshot(name)
            expiry = snapshot.get_created_at().shift(days=lock_days)
            if lock_days:
                log.uinfo("Setting snap lock to expire at %s" %
                          expiry.to('local'))
                snapshot.update_lock_expires_at(expiry)
        return snapshot
    except APICommandFailed as e:
        if e.error_code in ('NOT_SUPPORTED_IN_SYSTEM_STATE'):
            log.uabort('Operation not supported in STANDBY mode. Exiting.')
        else:
            exctype, excvalue, exctraceback = sys.exc_info()
            if hasattr(excvalue, 'message'):
                log.uwarn("Failed to create latest snapshot ({})".format(
                    excvalue.message))
            else:
                log.uwarn("Failed to create latest snapshot", exc_info=True)
            return None

    except Exception:
        exctype, excvalue, exctraceback = sys.exc_info()
        if hasattr(excvalue, 'message'):
            log.uwarn("Failed to create latest snapshot ({})".format(
                excvalue.message))
        else:
            log.uwarn("Failed to create latest snapshot", exc_info=True)
        return None
