"""

Usage:
   snaprotator [options]
   snaprotator -h | --help
   snaprotator --version


Options:
   -c configfile --configfile=<configfile>        Config file location
   -H hostname --hostname=<hostname>              DNS name of the infinibox
   -u user --user=<username>                      Administrator's username
   -p password --password=<password>              Administrator's password
   -v volume1[,volume2...] --volumes=<volume1[,volume2...>
                                                  Volume(s) to rotate the snaphsots on
   -b basename --basename=<basename>              Basename (daily,weekly, etc...)
   -r retain --retain=<retain>                    Number of snapshots to retain
   --lockdays=<snap_lock_days>                    Number of days to lock snapshot
   --maxlockdays=<max_snap_lock_days>             Max number of days a snap can be locked
   -l logfile --logfile=<logfile>                 Use log file instead of stderr
   -h, --help                                     Show this screen
   -D, --debug                                    Debug
   -q, --quiet                                    Suppress all informational messages
   -d, --dryrun                                   Show what will be done
   -P, --purge                                    Purge all snapshots with basename <basename>
   -s, --syslog                                   Log to syslog daemon
   -n, --norotation                               Do NOT rotate logfile, keep writing indefinitely
   -V, --version                                  Show snaprotator version
   -Z, --force_report                             Force heartbeat (last 24 hours report)
   -R, --delete                                   Delete all snaps with given basename using wildcard
   -L, --list                                     Show all snapshots belonging to policy
   -J, --listjson                                 Show all snapshots belonging to policy in JSON format
   -e password --encrypt=<password>               Encrypt cleartext password for config file use
"""
import signal
import sys
import logbook
from .logsetup import UserLogger, setup_logging, setup_mail_logging, has_warnings

from docopt import docopt
from .__version__ import __version__ as snaprotator_version
from .rotation import (create_latest_snapshot, trim_number_of_snapshots)

from . import vmware
from . import config as conf
from .infinibox import (
    do_metadata,
    connect_to_infinibox,
    infinibox_validate,
    volume_validate,
    get_all_volumes,
    get_all_volumes_by_serial,
    check_count_fs_snaps,
    get_all_filesystems,
    event_unknown_error,
    event_locked,
    get_all_snapshots
)
from infi.pyvmomi_wrapper import Client
from time import sleep, time
from . import powertools
from . import errors
from infi.pyutils.contexts import contextmanager
from collections import OrderedDict
import json
from shutil import copyfile
import os
import uuid

log = UserLogger(__name__)

MAX_ATTEMPTS = 3
RETRY_IN = 5
MAX_RETAIN = 1000
DEFAULT_PARALLEL = 5


def timed(func):
    def wrapper(*arg, **kw):
        t1 = time()
        res = func(*arg, **kw)
        t2 = time()
        took = t2 - t1
        tooks = int(took)
        tookms = int(took * 1000 - tooks)
        log.debug("{} took {}.{} seconds".format(func.__name__, tooks, tookms))
        return res
    return wrapper


@contextmanager
def locking_context(name):
    from lockfile import AlreadyLocked
    from lockfile.pidlockfile import PIDLockFile
    from os import path
    lock = PIDLockFile(name, timeout=0)

    while True:
        try:
            if lock.i_am_locking():  # our PID is already acquired the lock
                log.uinfo(
                    "snaprotator process already holds the pid lock file")
                break
            lock.acquire(0)
            log.debug("Lock acquired successfully")
            break  # successfully acquired the lock
        except AlreadyLocked:
            locking_pid = lock.read_pid()  # either None or int
            if locking_pid is not None and path.exists("/proc/{}".format(locking_pid)):
                log.uabort_with_exc_info("another snaprotator process (pid {}) is still running and did not release the lock file"
                                         .format(locking_pid), exit_code=errors.EXIT_CODE_ALREADY_RUNNING)
            # pid file is left over by a dead process, break the lock and
            # re-acquire
            log.uinfo("detected lock left over of a dead snaprotator process ({}), breaking lock".format(
                locking_pid))
            lock.break_lock()

    try:
        yield
    finally:
        lock.release()


def signal_handler(signal, frame):
    msg = (
        'Execution has been interrupted by Ctrl+C!\n' +
        'Please re-run the tool, to make sure the rotation is in order'
    )
    log.uwarn(msg)
    sys.exit(errors.EXIT_CODE_INTERRUPT)


@timed
def connect_to_vcenter(config):
    from requests.exceptions import (
        ConnectionError, ConnectTimeout, ReadTimeout, SSLError, ProxyError, RetryError)
    from pyVmomi import vim

    hostname = config['vcenter']['host']
    username = config['vcenter']['username']
    password = config['vcenter']['password']

    log.uinfo("Connecting to VCenter '{}'".format(hostname))
    try:
        client = Client(hostname, username=username, password=password)
        log.uinfo("vCenter version: {}".format(
            client.service_content.about.fullName))
    except (ConnectionError, ConnectTimeout, ReadTimeout, SSLError, ProxyError, RetryError) as e:
        log.uabort_with_exc_info("VCenter connection error, aborting. {}", str(e),
                                 exit_code=errors.EXIT_CODE_VMWARE_NETWORK_ERROR)
    except vim.fault.InvalidLogin as e:
        log.uabort_with_exc_info("VCenter login error, aborting. {}", e.msg,
                                 exit_code=errors.EXIT_CODE_VMWARE_LOGIN_FAILED)
    except SystemExit:
        raise
    except Exception as e:
        log.uabort_with_exc_info("Failed to connect to VCenter, aborting. Unknown error: {}".format(repr(e)),
                                 exit_code=errors.EXIT_CODE_VMWARE_GENERAL_ERROR)

    return client


def pipeline_vm_tasks(parallel, items, task_factory, task_handler):
    tasks = dict()
    while items or tasks:
        if len(tasks) < parallel and items:
            item = items.pop()
            task = task_factory(item)
            if task is not None:
                tasks[item] = task

        for item, task in tasks.copy().items():
            if task_handler(item, task):
                del tasks[item]
        sleep(0.1)


@timed
def do_infinibox_snapshot(infinibox, basename, retain, volume_name, dryrun,
                          uuid=False, suppress=False, delete=False, is_fs=False, lock_days=0, snap_both_aa_peers=True):
    remote_snap = None
    if not is_fs and snap_both_aa_peers:
        volumes = infinibox.objects.volumes.find(name=volume_name, type='MASTER').to_list()
        cgs = infinibox.objects.cons_groups.find(name=volume_name, type='MASTER').to_list()
        all_objects = volumes + cgs
        if all_objects:
            volume = all_objects[0]
            volume_type = volume.get_type_name()
            log.uinfo('Checking {} {}'.format(volume_type, volume.get_name()))
            is_active_active = False
            if volume.is_replicated():
                if volume_type == 'cons_group':
                    is_active_active = [a for a in volume.get_replicas(
                    ).to_list() if a.is_type_active_active()] is not []
                else:
                    is_active_active = volume.get_replica().is_type_active_active()

            if is_active_active:
                try:
                    log.uinfo("{} {} is ACTIVE_ACTIVE".format(
                        volume_type.capitalize(), volume.get_name()))
                    if volume_type == 'cons_group':
                        base = [a for a in volume.get_replicas(
                        ).to_list() if a.is_type_active_active()][0]
                    else:
                        base = volume.get_replica()

                    link_id = base.get_link().get_id()
                    res = infinibox.api.get(
                        "/api/rest/remote/{}/api/rest/config".format(link_id))
                    ip = res.get_json()['result']['ip_config']['ip']
                    from infinisdk import InfiniBox
                    remote_infinibox = InfiniBox(
                        ip, auth=infinibox.get_api_auth())
                    remote_infinibox.login()
                    infinibox.register_related_system(remote_infinibox)
                    remote_volume_name = base.get_remote_entity().get_name()
                    remote_snap, remote_metadata = _do_infinibox_snapshot(remote_infinibox, basename, retain, remote_volume_name, dryrun,
                                                                          uuid, suppress, delete, is_fs, lock_days)
                    if remote_snap:
                        log.uinfo('Successfully created remote snapshot on {}/{}'.format(
                            remote_infinibox.get_name(), remote_volume_name))
                except Exception as e:
                    log.uwarn(repr(e))
                    log.uwarn("Couldn't create remote snapshot. Continuing.")

    local_snap, metadata = _do_infinibox_snapshot(infinibox, basename, retain, volume_name, dryrun,
                                                  uuid, suppress, delete, is_fs, lock_days)
    if local_snap:
        log.uinfo('Successfully created local snapshot on {}/{}'.format(
                        infinibox.get_name(), volume_name))

    return local_snap, metadata, remote_snap


def _do_infinibox_snapshot(infinibox, basename, retain, volume_name, dryrun,
                           uuid=False, suppress=False, delete=False, is_fs=False, lock_days=0):
    volume, snapshots, too_many_snaps = volume_validate(infinibox, volume_name,
                                                        is_fs, True if lock_days else False)
    if not delete:
        if not too_many_snaps:
            snapshot_obj = create_latest_snapshot(
                infinibox, volume, basename, dryrun, uuid, suppress,
                lock_days)
        else:
            log.debug('Too many snaps, not creating a new one')
            snapshot_obj = None
        deleted_snaps_metadata = trim_number_of_snapshots(
            infinibox, snapshots, volume, basename, retain - 1, dryrun, suppress)
    else:
        snapshot_obj = None
        deleted_snaps_metadata = trim_number_of_snapshots(
            infinibox, snapshots, volume, basename, 0, dryrun, suppress)
    return snapshot_obj, deleted_snaps_metadata


@timed
def iter_filtered_datastores_volumes(client, config, infinibox, datastores, all_volumes):

    log.debug("iter_filtered_datastores_volumes")
    filtered_datastores = conf.list_of_datastores_to_snapshot(
        config, datastores)
    log.debug("conf.list_of_datastores_to_snapshot")
    infi_datastores = vmware.iter_filtered_datastores_block_devices(
        client, filtered_datastores)
    log.debug("vmware.iter_filtered_datastores_block_devices")
    for datastore, infinidat_devices in infi_datastores:
        for device in infinidat_devices:
            # volume = infinibox.objects.volumes.find(serial=device)
            # Doesn't work anymore with AA - ticket INFINIBOX-38512
            volume = all_volumes.get(device)
            if volume:
                yield datastore, datastore.name, datastore.vm, [volume]


def get_datastore_infinibox_volumes(client, config, infinibox, datastores, all_volumes):
    return list(iter_filtered_datastores_volumes(client, config, infinibox, datastores, all_volumes))


def get_datastore_snapshot(infinibox_snapshot, datastores_volumes, remote_snapshot):
    for datastore, datastore_name, datastore_vms, volume_list in datastores_volumes:
        if infinibox_snapshot.get_parent() in volume_list:
            return datastore, datastore_name, datastore_vms, infinibox_snapshot, remote_snapshot


def get_vms_infinisnapid_datastore(datastores_snapshot, processed_vms, consistent_only, include_templates):
    datastore, datastore_names, datastore_vms, infinisnap, remote_snap = datastores_snapshot
    vms = vmware.get_all_vms_on_datastore_with_single_datastore(
        datastore, processed_vms, consistent_only, include_templates)
    return vms, infinisnap, datastore, remote_snap


def get_vms_in_datastore_within_infinisnapid(datastore_volumes, infinibox_snapshot, datastores, vms, consistent_only, remote_snapshot, include_templates):
    datastore_snapshot = get_datastore_snapshot(
        infinibox_snapshot, datastore_volumes, remote_snapshot)
    log.debug("Finished get_datastore_snapshot")
    return get_vms_infinisnapid_datastore(datastore_snapshot, vms, consistent_only, include_templates)


@timed
def parallel_create_vmware_snapshots(client, parallel, vms, basename, guid):
    vmware_snapshots = []

    def create_snap_task_factory(vm_tuple):
        vm, quiesce, memory = vm_tuple
        return vmware.initiate_snapshot_for_vm(client, basename, vm, quiesce=quiesce, memory=memory, guid=guid)

    def create_snap_task_handler(vm_tuple, task):
        try:
            vm = vm_tuple[0]
            state, message = vmware.get_task_status(task)
            if state == 'success':  # success
                snapshot = vmware.wait_for_result(task)
                vmware_snapshots.append((vm, snapshot))
                log.uinfo(
                    'Successfully created snapshot for VM {} ({})', vm.name, snapshot)
            elif state == 'error':
                log.uwarn(
                    'Snapshot creation failed. VM Name: {}, reason: {}', vm.name, message)
        except Exception as e:
            log.uwarn('Unexpected error in create snap {}'.format(
                repr(e)), exc_info=True)
            state = 'error'
        return state in ('success', 'error')

    pipeline_vm_tasks(
        parallel, vms, create_snap_task_factory, create_snap_task_handler)
    return vmware_snapshots


@timed
def parallel_remove_vmware_snapshots(parallel, vmware_snapshots):
    def remove_snap_task_factory(vm_snapshot_pair):
        return vmware.initiate_remove_snapshot(vm_snapshot_pair)

    def remove_snap_task_handler(vm_snapshot_pair, task):
        try:
            vm, snapshot = vm_snapshot_pair
            state, message = vmware.get_task_status(task)
            if state == 'success':  # success
                log.uinfo(
                    'Successfully deleted snapshot {} for VM {}', snapshot, vm.name)
            elif state == 'error':
                log.uwarn(
                    'Snapshot deletion failed. VM Name: {}, reason: {}', vm.name, message)
        except Exception as e:
            log.uwarn('Unexpected error in remove snap {}'.format(
                repr(e)), exc_info=True)
            state = 'error'
        return state in ('success', 'error')

    pipeline_vm_tasks(
        parallel, vmware_snapshots, remove_snap_task_factory, remove_snap_task_handler)


@timed
def cleanup_vmware_snapshots(vms, client, basename):
    for vm, quiesce, memory in vms:
        try:
            log.debug('Cleanup VMWare snapshots for host {}', vm.name)
            vmware.cleanup_infinidat_snapshots(client, basename, vm)
        except Exception as e:
            log.uwarn('Failed to cleanup snapshot {}'.format(repr(e)))


@timed
def register_snap_unregister_snaps(
        powertools_info, vms_infinisnapid_datastore, metadata_list, guid):
    return powertools.register_snap_unregister_snaps(
        powertools_info, vms_infinisnapid_datastore, metadata_list, guid)


@timed
def do_infinibox_vmware_snapshots(client, config, parallel, retain, infinibox,
                                  basename, dryrun, suppress=False, delete=False, consistent_only=False,
                                  lock_days=0, snap_both_aa_peers=True, include_templates=False):
    infini_datastores = vmware.get_all_infinidat_datastores(client)
    all_volumes = get_all_volumes_by_serial(infinibox)
    datastores_to_snapshot = get_datastore_infinibox_volumes(
        client, config, infinibox, infini_datastores, all_volumes)
    if not datastores_to_snapshot:
        log.uabort("Can't find datastores to snapshot. Exiting.",
                   exit_code=errors.EXIT_CODE_VMWARE_NO_DATASTORES_FOUND)
    all_vms = list(vmware.iter_all_vms(client))
    log.info(all_vms)
    all_vms_including_powered_off = list(
        vmware.iter_all_vms(client, show_powered_off=True, include_templates=include_templates))
    use_volid = conf.is_use_volid(config)
    powertools_info = None
    todo_volumes = 0
    done_volumes = 0
    if not dryrun:
        powertools_info = powertools.get_powertools_info(client)
    datastore_names = [datastore_name for datastore, datastore_name,
                       datastore_vms, infinibox_volumes in datastores_to_snapshot]
    conf.report_missing_vms(config, datastore_names,
                            all_vms_including_powered_off)
    for datastore_volume_tuple in datastores_to_snapshot:
        # zip(*datastore_volume_tuple)
        datastore, datastore_name, datastore_vms, infinibox_volumes = datastore_volume_tuple
        # datastore, datastore.name, datastore.vm, volume
        # datastore_names = [datastore.name for datastore in datastores]
        log.uinfo('Datastore to snap: {}', datastore_name)
        guid = uuid.uuid4().hex

        vms_tuple = list(conf.iter_vms(config, [datastore_name], all_vms))
        vms = []

        if not vms_tuple:
            if "\n".join(list(config['vms'].keys())) == '!*':
                log.info(
                    'All VMS excluded. Continuing infinibox snapshots without OS consistency.')
            else:
                log.uwarn(
                    'No VMs found. Continuing infinibox snapshots without OS consistency.')
            vms_names = []
        else:
            vms_names = [vmname for vmname, vmdatastore,
                         vm in [vm for vm, quiesce, memory in vms_tuple]]
            for vmobj, quiesce, memory in vms_tuple:
                vmname, vmdatastore, vm = vmobj
                res = vm, quiesce, memory
                vms.append(res)
        log.uinfo('VMS to snap: {}', [str(a) for a in vms_names])
        log.debug("infinibox volumes: {}".format(
            [vol for vol in infinibox_volumes]))
        volume_names = [volume.get_name()
                        for vlist in [infinibox_volumes] for volume in vlist]
        todo_volumes += len(volume_names)

        log.uinfo(
            "Infinibox volumes to snap: {}", [str(a) for a in volume_names])
        vmware_snapshots = None
        if not dryrun and not delete:
            vmware_snapshots = parallel_create_vmware_snapshots(
                client, parallel, list(vms), basename, guid)

        try:
            for infinibox_volume in volume_names:
                try:
                    done_volumes += 1
                    snapshot_name, metadata_list, remote_snapshot = do_infinibox_snapshot(
                        infinibox, basename, retain, infinibox_volume, dryrun,
                        use_volid, suppress, delete, is_fs=False,
                        lock_days=lock_days, snap_both_aa_peers=snap_both_aa_peers)
                    if not snapshot_name:
                        done_volumes -= 1

                except SystemExit as e:
                    done_volumes -= 1
                    if e.code == errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND:
                        log.uwarn(
                            "Infinibox volume '{}' not found, skipping it.", infinibox_volume)
                        continue
                    else:
                        raise
                if not dryrun:
                    if powertools_info:
                        if delete:
                            vms_infinisnapid_datastore = None
                        else:
                            vms_infinisnapid_datastore = get_vms_in_datastore_within_infinisnapid(
                                datastores_to_snapshot, snapshot_name, infini_datastores, vms, consistent_only, remote_snapshot, include_templates)
                            log.debug('About to register the following {}'.format(
                                vms_infinisnapid_datastore[0][1]))

                        register_snap_unregister_snaps(
                            powertools_info, vms_infinisnapid_datastore, metadata_list, guid)
                        log.debug("Finished register/unregister")
                    else:
                        log.uinfo(
                            'No Host Power Tools For VMWare found. Check appliance installed/powered on.')
        finally:
            if not dryrun:
                parallel_remove_vmware_snapshots(parallel, vmware_snapshots)
                cleanup_vmware_snapshots(vms, client, basename)
    return todo_volumes, done_volumes


@timed
def do_purge(infinibox, basename, volume_name, dryrun, suppress=False, is_fs=False):
    volume, snapshots, too_many_snaps = volume_validate(
        infinibox, volume_name, is_fs)
    deleted_snaps_metadata = trim_number_of_snapshots(
        infinibox, snapshots, volume, basename, 0, dryrun, suppress)
    # fixme add vmware deletion
    return deleted_snaps_metadata


@timed
def wrapper_get_all_volumes(infinibox, allow_replicas=False):
    return get_all_volumes(infinibox, allow_replicas)


@timed
def wrapper_get_all_filesystems(infinibox, allow_replicas=False, snap_mapped_fs_snapshots=False):
    return get_all_filesystems(infinibox, allow_replicas, snap_mapped_fs_snapshots)


def copy_example_config():
    cwd = os.path.dirname(__file__)
    example_config = '{}/snaprotator.conf'.format(cwd)
    destination = '/etc/snaprotator.conf.example'
    try:
        copyfile(example_config, destination)
    except Exception:
        pass


def main():
    signal.signal(signal.SIGINT, signal_handler)
    arguments = docopt(__doc__, version=snaprotator_version)
    if arguments.get('--encrypt'):
        sys.stdout.write('{}\n'.format(conf.mask(arguments.get('--encrypt'))))
        sys.exit(0)
    config = conf.load_config(arguments) if arguments['--configfile'] else {}
    conf.update_arguments_to_config(arguments, config)
    if not config:
        docopt(__doc__, '-h')

    validate, error = conf.validate_config(config)
    if not validate:
        sys.stderr.write(
            'Configuration file or command line parameters error: {}\n'.format(error.errors))
        sys.exit(errors.EXIT_CODE_CONFIG_ERROR)

    log_handler = setup_logging(config)

    basename = config['global']['snap_basename']
    if 'snaps_to_keep' in config['global']:
        retain = config['global']['snaps_to_keep']
    else:
        retain = 0

    if 'snap_lock_days' in config['global']:
        lock_days = int(config['global']['snap_lock_days'])
    else:
        lock_days = 0

    if 'purge' in config['infinibox']:
        purge = config['infinibox']['purge']
    else:
        purge = False
    dryrun = conf.is_dryrun(config)
    force_report = conf.is_force_report(config)
    suppress = conf.is_suppress(config)
    is_vmware = conf.is_vmware(config)
    use_volid = conf.is_use_volid(config)
    single_instance = conf.is_single_instance(config)
    sa_internal = conf.is_sa_internal(config)
    consistent_only = conf.is_consistent_only(config)
    include_templates = conf.is_include_templates(config)

    delete = conf.is_delete(config)
    listplain = conf.is_list(config)
    listjson = conf.is_listjson(config)
    allow_replicas = conf.is_allow_replicas(config)
    snap_mapped_fs_snapshots = conf.is_snap_mapped_fs_snapshots(config)
    snap_both_aa_peers = conf.is_snap_both_aa_peers(config)
    mail_handler = setup_mail_logging(config, dryrun)
    infinibox = None
    done_vol_count = 0
    done_fs_count = 0
    todo_vol_count = 0
    todo_fs_count = 0
    standalone_volumes_filesystems = None
    standalone_volumes = None
    standalone_filesystems = None
    with logbook.NullHandler().applicationbound(), mail_handler.applicationbound():
        with log_handler.applicationbound():
            try:
                log.uinfo('SnapRotator v.{}'.format(snaprotator_version))
                if dryrun:
                    log.uinfo('DRYRUN')
                if not dryrun and delete:
                    log.uinfo(
                        'Deleting snapshots with basename {}'.format(basename))
                infinibox = connect_to_infinibox(config)
                infinibox_validate(infinibox)

                try:
                    with locking_context("/tmp/snaprotator.pid") if single_instance else conf.dummy_context_mgr():
                        if listjson or listplain:
                            snapshots = get_all_snapshots(
                                infinibox, allow_replicas, basename)
                            mydic = OrderedDict()

                            for snap in snapshots:
                                entry = dict()
                                name = snap.get_name()
                                entry['typename'] = typename = snap.get_type_name()
                                entry['parent'] = snap.get_parent().get_name()
                                entry['size'] = str(
                                    snap.get_size()) if typename != 'cons_group' else '<N/A>'
                                entry['used'] = str(
                                    snap.get_used_size()) if typename != 'cons_group' else '<N/A>'
                                mydic[name] = entry
                            # print mydic, using print to output to stdin
                            if listplain:
                                print(('{:30} {:10} {:10} {:10} {:28}'.format(
                                    'Name', 'Type', 'Size', 'Used', 'Parent')))
                                for key, val in mydic.items():
                                    print(('{:30} {:10} {:10} {:10} {:28}'.format(
                                        key, val['typename'], val['size'], val['used'], val['parent'])))
                            else:
                                print((json.dumps(mydic)))

                        else:  # Not listjson or listplain
                            # volumes key must exist because validate_config
                            # ensured that
                            if not is_vmware:
                                if 'volumes' in config['infinibox']:
                                    standalone_volumes_filesystems = config[
                                        'infinibox']['volumes'].split(',')
                                    log.uinfo('The following volumes will be snapshotted: {}', [
                                              str(a) for a in standalone_volumes_filesystems])
                                else:
                                    all_volume_objects = wrapper_get_all_volumes(
                                        infinibox, allow_replicas)
                                    all_volumes = [vol.get_name()
                                                   for vol in all_volume_objects]
                                    standalone_volumes = conf.list_of_volumes_to_snapshot(
                                        config, all_volumes)

                                    log.uinfo(
                                        'The following volumes will be snapshotted: {}', [str(a) for a in standalone_volumes])
                                    all_filesystem_objects = wrapper_get_all_filesystems(
                                        infinibox, allow_replicas, snap_mapped_fs_snapshots)

                                    all_filesystems = [fs.get_name()
                                                       for fs in all_filesystem_objects]
                                    standalone_filesystems = conf.list_of_filesystems_to_snapshot(
                                        config, all_filesystems)
                                    log.uinfo(
                                        'The following filesystems will be snapshotted: {}', [str(a) for a in standalone_filesystems])
                                if not standalone_filesystems and not standalone_volumes and not standalone_volumes_filesystems:
                                    if sa_internal:
                                        log.uwarn(
                                            "No filesystems/volumes/cgs found.")
                                        raise SystemExit(
                                            errors.EXIT_CODE_NOTHING_TODO)
                                    else:
                                        log.uabort_with_exc_info(
                                            "No filesystems/volumes/cgs found. Aborting.")
                                if not standalone_volumes_filesystems:
                                    todo_fs_count = len(standalone_filesystems)
                                    todo_vol_count = len(standalone_volumes)
                                    log.debug(
                                        "Total of {} fs, {} volumes to snap".format(todo_fs_count, todo_vol_count))
                                else:
                                    todo_vol_fs_count = len(
                                        standalone_volumes_filesystems)
                                    log.debug(
                                        'Total vol/fs/cg: {}'.format(todo_vol_fs_count))

                            if purge:
                                if not standalone_volumes_filesystems:
                                    for volume in standalone_volumes:
                                        do_purge(infinibox, basename,
                                                 volume, dryrun, is_fs=False)
                                    for filesystem in standalone_filesystems:
                                        do_purge(
                                            infinibox, basename, filesystem, dryrun, is_fs=True)
                                else:
                                    for volume in standalone_volumes_filesystems:
                                        do_purge(infinibox, basename,
                                                 volume, dryrun, is_fs='unknown')
                            elif is_vmware:
                                try:
                                    client = connect_to_vcenter(config)
                                    parallel = config['vcenter']['parallel'] if 'parallel' in config[
                                        'vcenter'] else DEFAULT_PARALLEL
                                    todo_vol_count, done_vol_count = do_infinibox_vmware_snapshots(
                                        client, config, parallel, retain,
                                        infinibox, basename, dryrun, suppress,
                                        delete, consistent_only, lock_days, snap_both_aa_peers, include_templates)
                                # Close the Vcenter connection
                                finally:
                                    try:
                                        client.logout()
                                    except Exception as e:
                                        pass
                            else:  # infinibox only
                                if not standalone_volumes_filesystems:
                                    done_vol_count = todo_vol_count
                                    for volume in standalone_volumes:
                                        try:
                                            snapshot_name, metadata_list, remote_snapshot = do_infinibox_snapshot(
                                                infinibox, basename, retain,
                                                volume, dryrun, use_volid,
                                                suppress, delete, is_fs=False,
                                                lock_days=lock_days, snap_both_aa_peers=snap_both_aa_peers)
                                            if not snapshot_name:
                                                done_vol_count -= 1
                                        except SystemExit as e:
                                            done_vol_count -= 1
                                            if e.code == errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND:
                                                log.uwarn(
                                                    'Infinibox volume {} not found, skipping it.', volume)
                                                continue
                                            else:
                                                raise
                                    done_fs_count = todo_fs_count
                                    for filesystem in standalone_filesystems:
                                        try:
                                            snapshot_name, metadata_list, remote_snapshot = do_infinibox_snapshot(
                                                infinibox, basename, retain,
                                                filesystem, dryrun, use_volid,
                                                suppress, delete, is_fs=True,
                                                lock_days=lock_days, snap_both_aa_peers=snap_both_aa_peers)
                                            if not snapshot_name:
                                                done_fs_count -= 1
                                        except SystemExit as e:
                                            done_fs_count -= 1
                                            if e.code == errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND:
                                                log.uwarn(
                                                    'Infinibox filesystem {} not found, skipping it.', filesystem)
                                                continue
                                            else:
                                                raise
                                else:
                                    done_vol_fs_count = todo_vol_fs_count
                                    for volume in standalone_volumes_filesystems:
                                        snapshot_name, metadata_list, remote_snapshot = do_infinibox_snapshot(
                                            infinibox, basename, retain,
                                            volume, dryrun, use_volid,
                                            suppress, delete, is_fs='unknown',
                                            lock_days=lock_days, snap_both_aa_peers=snap_both_aa_peers)
                                        if not snapshot_name:
                                            done_vol_fs_count -= 1
                            if has_warnings():
                                if not standalone_volumes_filesystems:
                                    log.uinfo('Done (with warnings). FS: {} out of {}, Vol: {} out of {}'.format(
                                        done_fs_count, todo_fs_count, done_vol_count, todo_vol_count))
                                else:
                                    log.uinfo('Done (with warnings). FS/CG/VOL: {} out of {}'.format(
                                        done_vol_fs_count, todo_vol_fs_count))
                                raise SystemExit(
                                    errors.EXIT_CODE_WARNINGS_FOUND)
                            else:
                                if not standalone_volumes_filesystems:
                                    log.uinfo('Finished without warnings or errors. FS: {} out of {}, Vol: {} out of {}'.format(
                                        done_fs_count, todo_fs_count, done_vol_count, todo_vol_count))
                                else:
                                    log.uinfo('Finished without warnings or errors. FS/CG/VOL: {} out of {}'.format(
                                        done_vol_fs_count, todo_vol_fs_count))
                                if not dryrun and not standalone_volumes_filesystems:
                                    do_metadata(infinibox, basename, done_fs_count, todo_fs_count -
                                                done_fs_count, done_vol_count, todo_vol_count - done_vol_count, 0, suppress, sa_internal, force_report)
                        log.info("Done.")
                except SystemExit as e:
                    if(e.code == errors.EXIT_CODE_ALREADY_RUNNING):
                        raise SystemExit(errors.EXIT_CODE_ALREADY_RUNNING)
                    elif(e.code == errors.EXIT_CODE_NOTHING_TODO):
                        raise SystemExit(errors.EXIT_CODE_NOTHING_TODO)
                    elif(e.code == errors.EXIT_CODE_WARNINGS_FOUND):
                        raise SystemExit(errors.EXIT_CODE_WARNINGS_FOUND)
                    else:
                        log.uwarn("{}".format(repr(e)))
                        raise SystemExit(errors.EXIT_CODE_GENERAL_ERROR)
                except Exception as e:
                    log.uwarn("{}".format(repr(e)))
                    raise SystemExit(errors.EXIT_CODE_GENERAL_ERROR)

            except SystemExit as e:
                if not dryrun:
                    if (e.code == errors.EXIT_CODE_WARNINGS_FOUND):
                        if not standalone_volumes_filesystems:
                            log.debug(
                                "Finished with warnings... Doing metadata.")
                            do_metadata(infinibox, basename, done_fs_count, todo_fs_count -
                                        done_fs_count, done_vol_count, todo_vol_count - done_vol_count, 1, suppress, sa_internal, force_report)
                        log.uwarn("Done with warnings.")
                    elif (e.code == errors.EXIT_CODE_NOTHING_TODO):
                        log.debug(
                            "No fs/cg/vol found, but its ok - sa_internal")
                        if not standalone_volumes_filesystems:
                            do_metadata(infinibox, basename, done_fs_count, todo_fs_count -
                                        done_fs_count, done_vol_count, todo_vol_count - done_vol_count, 0, suppress, sa_internal, force_report)
                        log.uinfo("Done. Nothing to do.")
                    elif (e.code == errors.EXIT_CODE_ALREADY_RUNNING):
                        event_locked(infinibox, sa_internal)
                        log.uwarn("Done with error(s).")
                    else:
                        event_unknown_error(infinibox, sa_internal)
                        log.uwarn("Done with error(s).")
                sys.exit(e.code)
            except Exception as e:
                log.uwarn("{}".format(repr(e)))
                log.debug("Unknown Error")
                if not dryrun:
                    event_unknown_error(infinibox, sa_internal)
                log.uabort_with_exc_info(
                    "Internal error occurred, please contact Infinidat support.")
